/* -*-  Mode: C++; c-file-style: "gnu"; indent-tabs-mode:nil; -*- */
//
// Copyright (c) 2009 Universita' di Firenze
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License version 2 as
// published by the Free Software Foundation;
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Author: Tommaso Pecorella <tommaso.pecorella@unifi.it>
//

#ifndef __wimax_1idp_application_h__
#define __wimax_1idp_application_h__

#include "ns3/address.h"
#include "ns3/application.h"
#include "ns3/event-id.h"
#include "ns3/ptr.h"
#include "ns3/data-rate.h"
#include "ns3/random-variable.h"
#include "ns3/traced-callback.h"

namespace ns3 {

class Address;
class RandomVariable;
class Socket;

/**
* \ingroup wimax
*
* \brief Generate traffic to a single destination according to an
*        IDP pattern.
*
* The model is based on an Interrupted Deterministic Process (IDP).
* One IDP represents one packet voice source.
 * Each IDP generates traffic between the hub and the subscriber unit.
*
* This traffic generator follows an On/Off pattern: after Application::StartApplication
* is called, "On" and "Off" states alternate. The duration of each of
* these states follow an Exponential distribution.
* During the "Off" state, no traffic is generated.
* During the "On" state, traffic cbr is generated. This traffic is
* characterized by the specified "data rate" and "packet size".
 *
 * Please read "Traffic Model for 802.16 TG3 MAC/PHY Simulations"
 * http://www.ieee802.org/16/tg3/contrib/802163c-01_30r1.pdf
 * for further details.
 */
class Wimax1idpApplication : public Application
{
public:
  static TypeId GetTypeId (void);

  Wimax1idpApplication ();

  virtual ~Wimax1idpApplication ();

  void SetMaxBytes (uint32_t maxBytes);

protected:
  virtual void DoDispose (void);
private:
  // inherited from Application base class.
  virtual void StartApplication (void);    // Called at time specified by Start
  virtual void StopApplication (void);     // Called at time specified by Stop

  // helpers
  void CancelEvents ();

  void Construct (Ptr<Node> n,
                  const Address &remote,
                  std::string tid,
                  uint32_t size);


  // Event handlers
  void StartSending ();
  void StopSending ();
  void SendPacket ();

  Ptr<Socket>     m_socket;       // Associated socket
  Address         m_peer;         // Peer address
  bool            m_connected;    // True if connected
  ExponentialVariable  m_onTime;       // rng for On Time
  ExponentialVariable  m_offTime;      // rng for Off Time
  ConstantVariable     m_pktTime;      // rng for Packet Time
  double          m_basicOnRate;  // basic On rate, to be rescaled
  double          m_basicOffRate; // basic Off rate, to be rescaled
  double          m_basicPktRate; // basic packet rate, to be rescaled
  uint32_t        m_pktSize;      // Size of packets
  uint32_t        m_maxBytes;     // Limit total number of bytes sent
  uint32_t        m_totBytes;     // Total bytes sent so far
  EventId         m_startStopEvent;     // Event id for next start or stop event
  EventId         m_sendEvent;    // Eventid of pending "send packet" event
  bool            m_sending;      // True if currently in sending state
  TypeId          m_tid;
  TracedCallback<Ptr<const Packet> > m_txTrace;

private:
  void ScheduleNextTx ();
  void ScheduleStartEvent ();
  void ScheduleStopEvent ();
  void ConnectionSucceeded (Ptr<Socket>);
  void ConnectionFailed (Ptr<Socket>);
  void Ignore (Ptr<Socket>);
};

} // namespace ns3

#endif

