/* -*-  Mode: C++; c-file-style: "gnu"; indent-tabs-mode:nil; -*- */
//
// Copyright (c) 2009 Universita' di Firenze
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License version 2 as
// published by the Free Software Foundation;
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Author: Tommaso Pecorella <tommaso.pecorella@unifi.it>
//

#include "ns3/log.h"
#include "ns3/address.h"
#include "ns3/node.h"
#include "ns3/nstime.h"
#include "ns3/data-rate.h"
#include "ns3/random-variable.h"
#include "ns3/socket.h"
#include "ns3/simulator.h"
#include "ns3/socket-factory.h"
#include "ns3/packet.h"
#include "ns3/uinteger.h"
#include "ns3/trace-source-accessor.h"
#include "wimax2irp-application.h"
#include "ns3/udp-socket-factory.h"

NS_LOG_COMPONENT_DEFINE ("Wimax2irpApplication");

using namespace std;

namespace ns3 {

NS_OBJECT_ENSURE_REGISTERED (Wimax2irpApplication);

TypeId
Wimax2irpApplication::GetTypeId (void)
{
  static TypeId tid = TypeId ("ns3::Wimax2irpApplication")
    .SetParent<Application> ()
    .AddConstructor<Wimax2irpApplication> ()
    .AddAttribute ("DataRate", "The data rate in on state.",
                   DataRateValue (DataRate ("1.9Mb/s")),
                   MakeDataRateAccessor (&Wimax2irpApplication::m_rate),
                   MakeDataRateChecker ())
    .AddAttribute ("Remote", "The address of the destination",
                   AddressValue (),
                   MakeAddressAccessor (&Wimax2irpApplication::m_peer),
                   MakeAddressChecker ())
    .AddAttribute ("MaxBytes",
                   "The total number of bytes to send. Once these bytes are sent, "
                   "no packet is sent again, even in on state. The value zero means "
                   "that there is no limit.",
                   UintegerValue (0),
                   MakeUintegerAccessor (&Wimax2irpApplication::m_maxBytes),
                   MakeUintegerChecker<uint32_t> ())
    .AddAttribute ("Protocol", "The type of protocol to use.",
                   TypeIdValue (UdpSocketFactory::GetTypeId ()),
                   MakeTypeIdAccessor (&Wimax2irpApplication::m_tid),
                   MakeTypeIdChecker ())
    .AddTraceSource ("Tx", "A new packet is created and is sent",
                     MakeTraceSourceAccessor (&Wimax2irpApplication::m_txTrace))
  ;
  return tid;
}


Wimax2irpApplication::Wimax2irpApplication ()
{
  NS_LOG_FUNCTION_NOARGS ();
  m_socket = 0;
  m_connected = false;
  m_totBytes = 0;
  m_pktSize = 188;
  m_basicOnRate[0] = 1.14;
  m_basicOnRate[1] = 1.54;
  m_basicOffRate[0] = 1.22;
  m_basicOffRate[1] = 1.28;
  m_basicPktRate[0] = 44.95;
  m_basicPktRate[1] = 61.90;
}

Wimax2irpApplication::~Wimax2irpApplication ()
{
  NS_LOG_FUNCTION_NOARGS ();
}

void
Wimax2irpApplication::SetMaxBytes (uint32_t maxBytes)
{
  NS_LOG_FUNCTION (this << maxBytes);
  m_maxBytes = maxBytes;
}


void
Wimax2irpApplication::DoDispose (void)
{
  NS_LOG_FUNCTION_NOARGS ();

  m_socket = 0;
  // chain up
  Application::DoDispose ();
}

// Application Methods
void Wimax2irpApplication::StartApplication () // Called at time specified by Start
{
  NS_LOG_FUNCTION_NOARGS ();

  // Create the socket if not already
  if (!m_socket)
    {
      m_socket = Socket::CreateSocket (GetNode (), m_tid);
      m_socket->Bind ();
      m_socket->Connect (m_peer);
    }
  // Insure no pending event and start them
  for (int i = 0; i < 2; i++ )
    {
      CancelEvents (i);
      double scaleFactor = (50.52 * m_pktSize * 8.0 ) / static_cast<double> (m_rate.GetBitRate ());
      m_onTime[i] = ParetoVariable (m_basicOnRate[i] / (m_basicOnRate[i] - 1), m_basicOnRate[i]);
      m_offTime[i] = ParetoVariable (m_basicOffRate[i] / (m_basicOffRate[i] - 1), m_basicOffRate[i]);
      m_pktTime[i] = ExponentialVariable (scaleFactor / m_basicPktRate[i]);
      ScheduleStartEvent (i);
      // std::cout << "m_onTime[" << i << "]:  " << scaleFactor/m_basicOnRate[i] << " or " << m_basicOnRate[i]/scaleFactor << std::endl;
      // std::cout << "m_offTime[" << i << "]: " << scaleFactor/m_basicOffRate[i] << " or " << m_basicOffRate[i]/scaleFactor << std::endl;
      // std::cout << "m_pktTime[" << i << "]: " << scaleFactor/m_basicPktRate[i] << " or " << m_basicPktRate[i]/scaleFactor << std::endl;
      // std::cout << "m_basicPktRate[" << i << "]: " << m_basicPktRate[i] << std::endl;
    }
}

void Wimax2irpApplication::StopApplication () // Called at time specified by Stop
{
  NS_LOG_FUNCTION_NOARGS ();

  for (int i = 0; i < 2; i++ )
    {
      CancelEvents (i);
    }
  if (m_socket != 0)
    {
      m_socket->Close ();
    }
  else
    {
      NS_LOG_WARN ("Wimax2irpApplication found null socket to close in StopApplication");
    }
}

void Wimax2irpApplication::CancelEvents (int genID)
{
  NS_LOG_FUNCTION_NOARGS ();

  // Cancel the pending send packet event
  Simulator::Cancel (m_sendEvent[genID]);
  Simulator::Cancel (m_startStopEvent[genID]);
}

// Event handlers
void Wimax2irpApplication::StartSending (int genID)
{
  NS_LOG_FUNCTION_NOARGS ();
  SendPacket (genID);  // Send a packet and schedule the send packet event
  ScheduleStopEvent (genID);
}

void Wimax2irpApplication::StopSending (int genID)
{
  NS_LOG_FUNCTION_NOARGS ();
  CancelEvents (genID);

  ScheduleStartEvent (genID);
}

// Private helpers
void Wimax2irpApplication::ScheduleNextTx (int genID)
{
  NS_LOG_FUNCTION_NOARGS ();

  if (m_maxBytes == 0 || m_totBytes < m_maxBytes)
    {
      uint32_t bits = m_pktSize * 8;
      NS_LOG_LOGIC ("bits = " << bits);
      Time nextTime = Seconds (m_pktTime[genID].GetValue ()); // Time till next packet
      NS_LOG_LOGIC ("nextTime = " << nextTime);
      m_sendEvent[genID] = Simulator::Schedule (nextTime,
                                                &Wimax2irpApplication::SendPacket, this, genID);
    }
  else
    { // All done, cancel any pending events
      StopApplication ();
    }
}

void Wimax2irpApplication::ScheduleStartEvent (int genID)
{  // Schedules the event to start sending data (switch to the "On" state)
  NS_LOG_FUNCTION_NOARGS ();

  Time offInterval = Seconds (m_offTime[genID].GetValue ());
  NS_LOG_LOGIC ("start at " << offInterval);
  m_startStopEvent[genID] = Simulator::Schedule (offInterval, &Wimax2irpApplication::StartSending, this, genID);
}

void Wimax2irpApplication::ScheduleStopEvent (int genID)
{  // Schedules the event to stop sending data (switch to "Off" state)
  NS_LOG_FUNCTION_NOARGS ();

  Time onInterval = Seconds (m_onTime[genID].GetValue ());
  NS_LOG_LOGIC ("stop at " << onInterval);
  m_startStopEvent[genID] = Simulator::Schedule (onInterval, &Wimax2irpApplication::StopSending, this, genID);
}


void Wimax2irpApplication::SendPacket (int genID)
{
  NS_LOG_FUNCTION_NOARGS ();
  NS_LOG_LOGIC ("sending packet at " << Simulator::Now ());
  NS_ASSERT (m_sendEvent[genID].IsExpired ());
  Ptr<Packet> packet = Create<Packet> (m_pktSize);
  m_txTrace (packet);
  m_socket->Send (packet);
  m_totBytes += m_pktSize;
  ScheduleNextTx (genID);
}

void Wimax2irpApplication::ConnectionSucceeded (Ptr<Socket>)
{
  NS_LOG_FUNCTION_NOARGS ();

  m_connected = true;
  for (int i = 0; i < 2; i++ )
    {
      ScheduleStartEvent (i);
    }
}

void Wimax2irpApplication::ConnectionFailed (Ptr<Socket>)
{
  NS_LOG_FUNCTION_NOARGS ();
  cout << "Wimax2irpApplication, Connection Failed" << endl;
}

} // Namespace ns3
